#lang cur

(require
 (only-in cur [new-elim elim]))

(define-datatype Nat : Type
  (z : Nat)
  (s : (-> Nat Nat)))

(define sub1
  (lambda (x : Nat)
    (elim
     x
     (lambda (x : Nat) Nat)
     z
     (lambda (n : Nat) (IH : Nat) n))))

;; A native Racket library.
(require
 (only-in rackunit check-equal?))

(check-equal?
 (sub1 z)
 z)

(check-equal?
 (sub1 (s z))
 z)

(check-equal?
 (sub1 (s (s z)))
 (s z))

#;(check-equal?
   (sub1 z)
   (s z))

(define plus
  (lambda (n : Nat) (m : Nat)
          (elim n
                (lambda (x : Nat) Nat)
                m
                (lambda (n-1 : Nat) (IH : Nat)
                        (s IH)))))

(check-equal?
 (plus z z)
 z)

(check-equal?
 (plus z (s z))
 (s z))

(check-equal?
 (plus (s z) (s z))
 (s (s z)))

(check-equal?
 (plus (s z) (s (s z)))
 (s (s (s z))))

;; Move everything to test submodule
(module+ test
  (check-equal?
   (plus z z)
   z)

  (check-equal?
   (plus z (s z))
   (s z))

  (check-equal?
   (plus (s z) (s z))
   (s (s z)))

  (check-equal?
   (plus (s z) (s (s z)))
   (s (s (s z)))))

;; Let's do everyone's favorite dependent type
#|
Run these in REPL due to bug in emacs-mode?

(define-datatype Vec (A : Type) : (-> Nat Type)
  (nil : (Vec A 0))
  (cons : (forall (a : A) (n : Nat) (Vec A n)
              (Vec A (s n)))))

(nil Nat)
(cons Nat z z (nil Nat))
(cons (Nat) (z) (z) (nil (Nat)))
(cons Nat z (s z) (nil Nat))

|#

;; Define, but not actually used

(define-datatype = (A : Type) (a : A) : (-> (b : A) Type)
  (refl : (= A a a)))

(refl Nat z)



(require
 cur/ntac/base
 cur/ntac/standard
 cur/stdlib/equality
 cur/ntac/rewrite)

#;(ntac (== Nat (plus (s z) (s z)) (s (s z)))
      ;interactive
      ;reflexivity
      ;interactive
       ;(by-exact (refl Nat (s (s z))))
        )

#;(ntac (forall (n : Nat)
              (== Nat (plus n z) n))
      interactive
      ;by-intros
      ;(by-induction n)
      ;reflexivity
      ;(by-rewrite IH5)
      ;reflexivity
      )

;; Okay, you get the picture. It's basically a proof assistant.

;; But what about the meta-language?

;; I don't know about you, but I'm getting tired of writing in unary.

;; Step 2
(require
 (only-in cur [#%datum super.datum]))

;; Step 3
(begin-for-syntax
  (define (nat->unary stx n)
    (let loop ([n n])
      (if (zero? n)
          (syntax/loc stx z)
          (quasisyntax/loc stx
            (s #,(loop (sub1 n))))))))

;; Step 1
(define-syntax #%datum
  ;; Now in the meta-language
  (lambda (stx)
    (syntax-parse stx
      [(_ . n:nat)
       (nat->unary stx (syntax->datum #'n))]
      [(_ . e)
       (quasisyntax/loc stx
         (super.datum . e))])))

(module+ test
  (check-equal? 0 z)
  (check-equal? 1 (s z))
  (check-equal? 2 (s (s z))))

(define-syntax :type-of
  (lambda (stx)
    (syntax-parse stx
      [(_ e)
       (printf "Type of ~a is ~a~n" (curnel->datum #'e)
               (curnel->datum (cur-type-infer #'e)))
       #'e])))

(:type-of 0)
(:type-of (refl Nat 0))
