#lang cur
(require
 (only-in cur [new-elim elim]))

;; (remember to switch the display window)

;; Now I want to work a more extended example.
;; Suppose you don't like my tactic language, and you'd rather write your own.
;; Okay, that's fine.
;; The tactic language is entirely a user-defined library.
;; It's basically just the following:

(define-syntax tac
  (lambda (stx)
    (syntax-parse stx
      [(_ goal cmds ...)
       (run-tac-interpreter #'goal
        (make-initial-state (cur-expand #'goal))
        (syntax->list #'(cmds ...)))])))

;; In fact, let's do it.
(begin-for-syntax
  ;; Step 1
  (struct proof-state (env focus goals term))
  (define (make-initial-state goal)
    (proof-state '() 0 (list goal) (lambda (x) x)))

  (define (run-tac-interpreter orig-goal ps cmds)
    ;; Step 2 (stop here, read in the rest of the code)
    (let ([final-ps
           (for/fold ([ps ps])
                     ([cmd cmds])
             (pretty-print-ps ps)
             ((eval-syntax cmd) ps))])
      (let ([final-term (proof-state-term final-ps)])
        (displayln (curnel->datum (cur-type-infer final-term)))
        (displayln (curnel->datum final-term))
        #;final-term
        (if (cur-type-check? final-term orig-goal)
            final-term
            (raise-syntax-error 'tac "Produced bad proof term"
                                final-term orig-goal)))))

  ;; Step 6
  (define (pretty-print-ps ps)
    (for ([(k v) (in-dict (proof-state-env ps))])
      (printf "~a: ~a~n" (curnel->datum k)
              (curnel->datum v)))
    (printf "------------------------------~n")
    (printf "~a~n~n" (curnel->datum (list-ref (proof-state-goals ps)
                                              (proof-state-focus ps)))))

  ;; Step 5
  (define (list-remove ls pos)
    (if (zero? pos)
        (cdr ls)
        (cons (car ls) (list-remove (sub1 pos) (cdr ls)))))

  ;; Step 4
  (define (by-assumption ps)
    (let ([goals (proof-state-goals ps)]
          [focus (proof-state-focus ps)])
      (for/or ([(k v) (in-dict (proof-state-env ps))])
        (if (cur-equal? v (list-ref goals focus))
            (struct-copy proof-state ps
                         [focus (if (= 1 (length goals))
                                    #f
                                    (modulo (sub1 focus) (length goals)))]
                         [goals (list-remove goals focus)]
                         [term ((proof-state-term ps)
                                k)])
            #f))))

  ;; Step 3
  (require racket/dict racket/list)
  (define (intro ps)
    (let ([goals (proof-state-goals ps)]
          [focus (proof-state-focus ps)])
      (syntax-parse (list-ref goals focus)
        [(~Π (x : A) B)
         (struct-copy proof-state ps
                      [goals (list-set goals focus #'B)]
                      [env (dict-set (proof-state-env ps)
                                     #'x
                                     #'A)]
                      [term (lambda (y)
                              ((proof-state-term ps)
                               #`(lambda (x : A)
                                   #,y)))])]))))

#;(tac (forall (A : Type) (a : A) A)
     intro
     intro
     by-assumption)

#;(define id (tac (forall (A : Type) (a : A) A)
                intro
                intro
                by-assumption))

#;(define-datatype Bool : Type [true : Bool] [false : Bool])

#;(id Bool true)

;; You, the user, are empowered to write essentially arbitrary little surface
;; languages this way. New cool pattern matching language that compiles to
;; elimnators? BNF and inference rules parser?
;; New domain-specific tactic system? Translator to LaTeX? Let your imagination
;; run wild.
;;
